@@ -74,7 +74,7 @@ struct a2dp_setup_cb {
 	a2dp_stream_cb_t resume_cb;
 	a2dp_stream_cb_t suspend_cb;
 	void *user_data;
-	int id;
+	unsigned int id;
 };
 
 struct a2dp_setup {
@@ -302,19 +302,27 @@ static gboolean sbc_setconf_ind(struct avdtp *session,
 	/* Check bipool range */
 	for (codec_cap = NULL; caps; caps = g_slist_next(caps)) {
 		cap = caps->data;
-		if (cap->category == AVDTP_MEDIA_CODEC) {
-			codec_cap = (void *) cap->data;
-			if (codec_cap->media_codec_type == A2DP_CODEC_SBC) {
-				sbc_cap = (void *) codec_cap;
-				if (sbc_cap->min_bitpool < MIN_BITPOOL ||
+		if (cap->category != AVDTP_MEDIA_CODEC)
+			continue;
+
+		if (cap->length < sizeof(struct sbc_codec_cap))
+			continue;
+
+		codec_cap = (void *) cap->data;
+
+		if (codec_cap->media_codec_type != A2DP_CODEC_SBC)
+			continue;
+
+		sbc_cap = (void *) codec_cap;
+
+		if (sbc_cap->min_bitpool < MIN_BITPOOL ||
 					sbc_cap->max_bitpool > MAX_BITPOOL) {
-					*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-					*category = AVDTP_MEDIA_CODEC;
-					return FALSE;
-				}
-			}
-			break;
+			*err = AVDTP_UNSUPPORTED_CONFIGURATION;
+			*category = AVDTP_MEDIA_CODEC;
+			return FALSE;
 		}
+
+		break;
 	}
 
 	avdtp_stream_add_cb(session, stream, stream_state_changed, a2dp_sep);
@@ -1122,10 +1130,16 @@ proceed:
 
 	server = find_server(servers, src);
 	if (!server) {
+		int av_err;
+
 		server = g_new0(struct a2dp_server, 1);
 		if (!server)
 			return -ENOMEM;
-		avdtp_init(src, config);
+
+		av_err = avdtp_init(src, config);
+		if (av_err < 0)
+			return av_err;
+
 		bacpy(&server->src, src);
 		servers = g_slist_append(servers, server);
 	}
