@@ -534,7 +534,6 @@ static int found_device_req_name(struct btd_adapter *adapter)
 void hcid_dbus_inquiry_complete(bdaddr_t *local)
 {
 	struct btd_adapter *adapter;
-	const gchar *path;
 	int state;
 
 	adapter = manager_find_adapter(local);
@@ -543,8 +542,6 @@ void hcid_dbus_inquiry_complete(bdaddr_t *local)
 		return;
 	}
 
-	path = adapter_get_path(adapter);
-
 	/*
 	 * The following scenarios can happen:
 	 * 1. standard inquiry: always send discovery completed signal
@@ -869,20 +866,8 @@ void hcid_dbus_remote_name(bdaddr_t *local, bdaddr_t *peer, uint8_t status,
 				NULL);
 	}
 
-	if (device) {
-		char alias[248];
-		const char *dev_path = device_get_path(device);
-
-		emit_property_changed(connection, dev_path,
-						DEVICE_INTERFACE, "Name",
-						DBUS_TYPE_STRING, &name);
-
-		if (read_device_alias(srcaddr, dstaddr,
-					alias, sizeof(alias)) < 1)
-			emit_property_changed(connection, dev_path,
-						DEVICE_INTERFACE, "Alias",
-						DBUS_TYPE_STRING, &name);
-	}
+	if (device)
+		device_set_name(device, name);
 
 proceed:
 	/* remove from remote name request list */
@@ -1017,6 +1002,8 @@ int set_service_classes(int dd, const uint8_t *cls, uint8_t value)
 
 	dev_class = (value << 16) | (cls[1] << 8) | cls[0];
 
+	debug("Changing service classes to 0x%06x", dev_class);
+
 	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
 		int err = -errno;
 		error("Can't write class of device: %s (%d)",
@@ -1035,6 +1022,8 @@ int set_major_and_minor_class(int dd, const uint8_t *cls,
 	dev_class = (cls[2] << 16) | ((cls[1] & 0x20) << 8) |
 						((major & 0xdf) << 8) | minor;
 
+	debug("Changing major/minor class to 0x%06x", dev_class);
+
 	if (hci_write_class_of_dev(dd, dev_class, HCI_REQ_TIMEOUT) < 0) {
 		int err = -errno;
 		error("Can't write class of device: %s (%d)",
@@ -1170,10 +1159,9 @@ void hcid_dbus_write_class_complete(bdaddr_t *local)
 		return;
 	}
 
-	adapter_set_class(adapter, cls);
-	write_local_class(local, cls);
-
 	hci_close_dev(dd);
+
+	adapter_set_class(adapter, cls);
 }
 
 void hcid_dbus_write_simple_pairing_mode_complete(bdaddr_t *local)
